//
//  SbdPlayer.cpp
//

#include "SbdPlayer.h"
#include "SbdPlayerData.h"
#include <string>


namespace sbd
{

/**
 * definition
 */

static const ss_u32 DATA_ID = 0x42505353;
static const ss_u32 DATA_VERSION = 1;


/**
 * utilites
 */

static void splitPath(std::string& directoty, std::string& filename, std::string& extname, const std::string& path)
{
    std::string f = path;
    std::string d = "";
	std::string e = "";

    size_t pos = path.find_last_of("/");
	if (pos == std::string::npos) pos = path.find_last_of("\\");	// for win

    if (pos != std::string::npos)
    {
        d = path.substr(0, pos+1);
        f = path.substr(pos+1);
    }

	std::string::size_type epos = f.rfind(".");
	if (epos != std::string::npos)
	{
		e = f.substr(epos);
		f = f.substr(0, epos);
	}

	directoty = d;
	filename = f;
	extname = e;
}


std::string replaceExtension(const std::string& path, const std::string& newExt)
{
	std::string nNewExt = (newExt.at(0) == '.') ? newExt : "." + newExt;

	std::string npath(path);

	std::string::size_type pos = npath.rfind(".");
	if (pos != std::string::npos)
	{
		std::string old = npath.substr(pos);
		npath = npath.substr(0, pos);
	}
	
	npath += nNewExt;
	return npath;
}




/**
 * ToPointer
 */
class ToPointer
{
public:
	explicit ToPointer(const void* base)
		: _base(static_cast<const char*>(base)) {}
	
	const void* operator()(ss_offset offset) const
	{
		return (_base + offset);
	}

private:
	const char*	_base;
};


/**
 * DataArrayReader
 */
class DataArrayReader
{
public:
	DataArrayReader(const ss_u16* dataPtr)
		: _dataPtr(dataPtr)
	{}

	ss_u16 readU16() { return *_dataPtr++; }
	ss_s16 readS16() { return static_cast<ss_s16>(*_dataPtr++); }

	unsigned int readU32()
	{
		unsigned int l = readU16();
		unsigned int u = readU16();
		return static_cast<unsigned int>((u << 16) | l);
	}

	int readS32()
	{
		return static_cast<int>(readU32());
	}

	float readFloat()
	{
		union {
			float			f;
			unsigned int	i;
		} c;
		c.i = readU32();
		return c.f;
	}
	
	void readColor(cocos2d::ccColor4B& color)
	{
		unsigned int raw = readU32();
		color.a = static_cast<GLubyte>(raw >> 24);
		color.r = static_cast<GLubyte>(raw >> 16);
		color.g = static_cast<GLubyte>(raw >> 8);
		color.b = static_cast<GLubyte>(raw);
	}
	
	ss_offset readOffset()
	{
		return static_cast<ss_offset>(readS32());
	}

private:
	const ss_u16*	_dataPtr;
};


/**
 * TextureSetRef
 */
struct TextureSetRef : public cocos2d::CCObject
{
	cocos2d::CCArray textures;
};


/**
 * CellRef
 */
struct CellRef : public cocos2d::CCObject
{
	const Cell* cell;
	TextureSetRef* textureSetRef;
	cocos2d::CCRect rect;
};


/**
 * CellCache
 */
class CellCache : public cocos2d::CCObject
{
public:
	static CellCache* create(const ProjectData* data, const std::string& imageBaseDir, const std::string& replacedExt)
	{
		CellCache* obj = new CellCache();
		if (obj)
		{
			obj->init(data, imageBaseDir, replacedExt);
			obj->autorelease();
		}
		return obj;
	}

	CellRef* getReference(int index)
	{
		if (index < 0 || index >= _refs.count())
		{
			CCLOGERROR("Index out of range > %d", index);
			CC_ASSERT(0);
		}
		CellRef* ref = (CellRef*)_refs.objectAtIndex(index);
		return ref;
	}

	void changeCell(const std::string& cellName, const std::string& changeTextureName, int x, int y)
	{
		CellRef* ref = (CellRef*)_cellRefDic.objectForKey(cellName);
		if (ref == NULL) return;
		TextureSetRef* textureSetRef = (TextureSetRef*)_textureDic.objectForKey(changeTextureName);
		if (textureSetRef == NULL) return;
		
		ref->textureSetRef = textureSetRef;
		ref->rect.origin = cocos2d::CCPoint(x, y);
	}

protected:
	void init(const ProjectData* data, const std::string& imageBaseDir, const std::string& replacedExt)
	{
		CCAssert(data != NULL, "Invalid data");
		
		_data = data;
		_textures.removeAllObjects();
		_refs.removeAllObjects();
		
		ToPointer ptr(data);
		const Cell* cells = static_cast<const Cell*>(ptr(data->cells));

		for (int i = 0; i < data->numCells; i++)
		{
			const Cell* cell = &cells[i];
			const CellMap* cellMap = static_cast<const CellMap*>(ptr(cell->cellMap));
			
			if (cellMap->index >= _textures.count())
			{
				const char* imagePath = static_cast<const char*>(ptr(cellMap->imagePath));
				addTexture(imagePath, imageBaseDir, replacedExt);
			}
			
			CellRef* ref = new CellRef();
			ref->cell = cell;
			ref->textureSetRef = (TextureSetRef*)_textures.objectAtIndex(cellMap->index);
			ref->rect = cocos2d::CCRect(cell->x, cell->y, cell->width, cell->height);
			_refs.addObject(ref);

			const char* cellName = static_cast<const char*>(ptr(cell->name));
			_cellRefDic.setObject(ref, cellName);
		}
	}

	void addTexture(const std::string& imagePath, const std::string& imageBaseDir, const std::string& replacedExt)
	{
		std::string path = "";
		
		if (cocos2d::CCFileUtils::sharedFileUtils()->isAbsolutePath(imagePath))
		{
			// 絶対パスのときはそのまま扱う
			path = imagePath;
		}
		else
		{
			// 相対パスのときはimageBaseDirを付与する
			path.append(imageBaseDir);
			size_t pathLen = path.length();
			if (pathLen && path.at(pathLen-1) != '/' && path.at(pathLen-1) != '\\')
			{
				path.append("/");
			}
			path.append(imagePath);
		}
		
		if (replacedExt.size() > 0)
		{
			path = replaceExtension(path, replacedExt);
		}
		
		std::string directory;
		std::string filename;
		std::string extname;
		splitPath(directory, filename, extname, path);
		
		std::string basename = filename.substr(0, filename.length() - 1);

		cocos2d::CCTextureCache* texCache = cocos2d::CCTextureCache::sharedTextureCache();
		
		TextureSetRef* ref = new TextureSetRef();

		for (int i = 1; i <= 2; i++)
		{
			path = directory;
			path.append(basename);
			path.append(i == 1 ? "1" : "2");
			path.append(extname);
			
			cocos2d::CCTexture2D* tex = texCache->addImage(path.c_str());
			if (tex == NULL)
			{
				std::string msg = "Can't load image > " + path;
				CCAssert(tex != NULL, msg.c_str());
			}
			CCLOG("load: %s", path.c_str());
			
			ref->textures.addObject(tex);
		}
		
		_textures.addObject(ref);
		_textureDic.setObject(ref, filename);
	}

protected:
	const ProjectData*		_data;
	cocos2d::CCArray		_textures;
	cocos2d::CCArray		_refs;

	cocos2d::CCDictionary	_textureDic;
	cocos2d::CCDictionary	_cellRefDic;
};



/**
 * PartsInfo
 */
struct PartsInfo : public cocos2d::CCObject
{
	struct PartIndex : public cocos2d::CCObject
	{
		int index;
	};
	
	cocos2d::CCDictionary	indices;
};


/**
 * AnimeRef
 */
struct AnimeRef : public cocos2d::CCObject
{
	std::string				packName;
	std::string				animeName;
	const AnimationData*	animationData;
	const AnimePackData*	animePackData;

	PartsInfo*				partsInfo;

	AnimeRef(PartsInfo* partsInfo)
		: partsInfo(partsInfo)
	{
		partsInfo->retain();
	}

	~AnimeRef()
	{
		partsInfo->release();
	}
};


/**
 * AnimeCache
 */
class AnimeCache : public cocos2d::CCObject
{
public:
	static AnimeCache* create(const ProjectData* data)
	{
		AnimeCache* obj = new AnimeCache();
		if (obj)
		{
			obj->init(data);
			obj->autorelease();
		}
		return obj;
	}

	/**
	 * packNameとanimeNameを指定してAnimeRefを得る
	 */
	AnimeRef* getReference(const std::string& packName, const std::string& animeName)
	{
		std::string key = toPackAnimeKey(packName, animeName);
		AnimeRef* ref = (AnimeRef*)_dic.objectForKey(key);
		return ref;
	}

	/**
	 * animeNameのみ指定してAnimeRefを得る
	 */
	AnimeRef* getReference(const std::string& animeName)
	{
		AnimeRef* ref = (AnimeRef*)_dic.objectForKey(animeName);
		return ref;
	}
	
	void dump()
	{
//		for (auto key : _dic.keys())
//		{
//			CCLOG("%s", key.c_str());
//		}
	}

protected:
	void init(const ProjectData* data)
	{
		CCAssert(data != NULL, "Invalid data");
		
		ToPointer ptr(data);
		const AnimePackData* animePacks = static_cast<const AnimePackData*>(ptr(data->animePacks));

		for (int packIndex = 0; packIndex < data->numAnimePacks; packIndex++)
		{
			const AnimePackData* pack = &animePacks[packIndex];
			const AnimationData* animations = static_cast<const AnimationData*>(ptr(pack->animations));
			const char* packName = static_cast<const char*>(ptr(pack->name));
			
			const PartData* parts = static_cast<const PartData*>(ptr(pack->parts));
			PartsInfo* partsInfo = new PartsInfo();
			for (int partIndex = 0; partIndex < pack->numParts; partIndex++)
			{
				const char* partName = static_cast<const char*>(ptr(parts[partIndex].name));
				PartsInfo::PartIndex* info = new PartsInfo::PartIndex();
				info->index = partIndex;
				partsInfo->indices.setObject(info, partName);
			}

			for (int animeIndex = 0; animeIndex < pack->numAnimations; animeIndex++)
			{
				const AnimationData* anime = &animations[animeIndex];
				const char* animeName = static_cast<const char*>(ptr(anime->name));
				
				AnimeRef* ref = new AnimeRef(partsInfo);
				ref->packName = packName;
				ref->animeName = animeName;
				ref->animationData = anime;
				ref->animePackData = pack;

				// packName + animeNameでの登録
				std::string key = toPackAnimeKey(packName, animeName);
				CCLOG("anime key: %s", key.c_str());
				_dic.setObject(ref, key);

				// animeNameのみでの登録
				_dic.setObject(ref, animeName);
			}
		}
	}

	static std::string toPackAnimeKey(const std::string& packName, const std::string& animeName)
	{
		return cocos2d::CCString::createWithFormat("%s/%s", packName.c_str(), animeName.c_str())->m_sString;
	}

protected:
	cocos2d::CCDictionary	_dic;
};





/**
 * ResourceSet
 */
struct ResourceSet : public cocos2d::CCObject
{
	const ProjectData* data;
	bool isDataAutoRelease;
	CellCache* cellCache;
	AnimeCache* animeCache;

	virtual ~ResourceSet()
	{
		if (isDataAutoRelease) delete data;
	}
};


/**
 * ResourceManager
 */

static ResourceManager* defaultInstance = NULL;
const std::string ResourceManager::s_null;

ResourceManager* ResourceManager::getInstance()
{
	if (!defaultInstance)
	{
		defaultInstance = ResourceManager::create();
		defaultInstance->retain();
	}
	return defaultInstance;
}

ResourceManager::ResourceManager(void)
{
}

ResourceManager::~ResourceManager()
{
}

ResourceManager* ResourceManager::create()
{
	ResourceManager* obj = new ResourceManager();
	if (obj)
	{
		obj->autorelease();
	}
	return obj;
}

ResourceSet* ResourceManager::getData(const std::string& dataKey)
{
	ResourceSet* rs = (ResourceSet*)_dataDic.objectForKey(dataKey);
	return rs;
}

cocos2d::CCTexture2D* ResourceManager::getFirstTexture(const std::string& dataKey)
{
	ResourceSet* rs = (ResourceSet*)_dataDic.objectForKey(dataKey);
	CellRef* cellRef = rs->cellCache->getReference(0);
	cocos2d::CCTexture2D* tex = (cocos2d::CCTexture2D*)cellRef->textureSetRef->textures.objectAtIndex(0);
	return tex;
}

std::string ResourceManager::addData(const std::string& dataKey, const ProjectData* data, const std::string& imageBaseDir, const std::string& replacedExt)
{
    CCAssert(data != NULL, "Invalid data");
	CCAssert(data->dataId == DATA_ID, "Not data id matched");
	CCAssert(data->version == DATA_VERSION, "Version number of data does not match");
	
	// imageBaseDirの指定がないときコンバート時に指定されたパスを使用する
	std::string baseDir = imageBaseDir;
	if (imageBaseDir == s_null && data->imageBaseDir)
	{
		ToPointer ptr(data);
		const char* dir = static_cast<const char*>(ptr(data->imageBaseDir));
		baseDir = dir;
	}

	std::string ext = replacedExt;
	if (ext == s_null) ext = "";
	
	CellCache* cellCache = CellCache::create(data, baseDir, ext);
	cellCache->retain();
	
	AnimeCache* animeCache = AnimeCache::create(data);
	animeCache->retain();
	
	ResourceSet* rs = new ResourceSet();
	rs->data = data;
	rs->isDataAutoRelease = false;
	rs->cellCache = cellCache;
	rs->animeCache = animeCache;
	_dataDic.setObject(rs, dataKey);
	
	return dataKey;
}

std::string ResourceManager::addDataWithKey(const std::string& dataKey, const std::string& ssbpFilepath, const std::string& imageBaseDir, const std::string& replacedExt)
{
	std::string fullpath = cocos2d::CCFileUtils::sharedFileUtils()->fullPathForFilename(ssbpFilepath.c_str());

	unsigned long nSize = 0;
	void* loadData = cocos2d::CCFileUtils::sharedFileUtils()->getFileData(fullpath.c_str(), "rb", &nSize);
	if (loadData == NULL)
	{
		std::string msg = "Can't load project data > " + fullpath;
		CCAssert(loadData != NULL, msg.c_str());
	}
	
	const ProjectData* data = static_cast<const ProjectData*>(loadData);
	CCAssert(data->dataId == DATA_ID, "Not data id matched");
	CCAssert(data->version == DATA_VERSION, "Version number of data does not match");
	
	std::string baseDir = imageBaseDir;
	if (imageBaseDir == s_null)
	{
		// imageBaseDirの指定がないとき
		if (data->imageBaseDir)
		{
			// コンバート時に指定されたパスを使用する
			ToPointer ptr(data);
			const char* dir = static_cast<const char*>(ptr(data->imageBaseDir));
			baseDir = dir;
		}
		else
		{
			// プロジェクトファイルと同じディレクトリを指定する
			std::string directory;
			std::string filename;
			std::string extname;
			splitPath(directory, filename, extname, ssbpFilepath);
			baseDir = directory;
		}
		//CCLOG("imageBaseDir: %s", baseDir.c_str());
	}

	addData(dataKey, data, baseDir, replacedExt);
	
	// リソースが破棄されるとき一緒にロードしたデータも破棄する
	ResourceSet* rs = getData(dataKey);
	CCAssert(rs != NULL, "");
	rs->isDataAutoRelease = true;
	
	return dataKey;
}

std::string ResourceManager::addData(const std::string& ssbpFilepath, const std::string& imageBaseDir, const std::string& replacedExt)
{
	// ファイル名を取り出す
	std::string directory;
    std::string filename;
	std::string extname;
	splitPath(directory, filename, extname, ssbpFilepath);
	
	std::string dataKey = filename;
	
	return addDataWithKey(dataKey, ssbpFilepath, imageBaseDir, replacedExt);
}

void ResourceManager::removeData(const std::string& dataKey)
{
	_dataDic.removeObjectForKey(dataKey);
}

void ResourceManager::removeAllData()
{
	_dataDic.removeAllObjects();
}

void ResourceManager::changeCell(const std::string& dataKey, const std::string& cellName, const std::string& changeTextureName, int x, int y)
{
	ResourceSet* rs = (ResourceSet*)_dataDic.objectForKey(dataKey);
	rs->cellCache->changeCell(cellName, changeTextureName, x, y);
}






/**
 * CustomSprite
 */
class CustomSprite : public cocos2d::CCSprite
{
public:
	CustomSprite*	_parent;
	bool			_isPartVisibled;
	int				_textureSetNo;

public:
	CustomSprite();
	virtual ~CustomSprite();

	static CustomSprite* create();

	void initState()
	{
		_isPartVisibled = true;
		_textureSetNo = 0;
	}
};





/**
 * Player
 */

static const std::string s_nullString;

Player::Player(void)
	: _resman(NULL)
	, _currentRs(NULL)
	, _currentAnimeRef(NULL)
	, _useSpriteBatchNode(false)

	, _flippedX(false)
	, _flippedY(false)
	, _frameSkipEnabled(true)
	, _playingFrame(0.0f)
	, _step(1.0f)
	, _loop(0)
	, _loopCount(0)
	, _isPlaying(false)
	, _isPausing(false)
	, _prevDrawFrameNo(-1)
{
}

Player::~Player()
{
	this->unscheduleUpdate();
	releaseParts();
	releaseData();
	releaseResourceManager();
}

Player* Player::create(ResourceManager* resman)
{
	Player* obj = new Player();
	if (obj && obj->init())
	{
		obj->setResourceManager(resman);
		obj->autorelease();
		obj->scheduleUpdate();
		return obj;
	}
	CC_SAFE_DELETE(obj);
	return NULL;
}

bool Player::init()
{
	//if (!cocos2d::CCSprite::init())
	if (!cocos2d::CCNode::init())
    {
        return false;
    }
	return true;
}

void Player::releaseResourceManager()
{
	CC_SAFE_RELEASE_NULL(_resman);
}

void Player::setResourceManager(ResourceManager* resman)
{
	if (_resman) releaseResourceManager();
	
	if (!resman)
	{
		// nullのときはデフォルトを使用する
		resman = ResourceManager::getInstance();
	}
	
	CC_SAFE_RETAIN(resman);
	_resman = resman;
}

int Player::getFrameNo() const
{
	return static_cast<int>(_playingFrame);
}

void Player::setFrameNo(int frameNo)
{
	_playingFrame = frameNo;
}

float Player::getStep() const
{
	return _step;
}

void Player::setStep(float step)
{
	_step = step;
}

int Player::getLoop() const
{
	return _loop;
}

void Player::setLoop(int loop)
{
	if (loop < 0) return;
	_loop = loop;
}

int Player::getLoopCount() const
{
	return _loopCount;
}

void Player::clearLoopCount()
{
	_loopCount = 0;
}

void Player::setFrameSkipEnabled(bool enabled)
{
	_frameSkipEnabled = enabled;
	_playingFrame = (int)_playingFrame;
}

bool Player::isFrameSkipEnabled() const
{
	return _frameSkipEnabled;
}


void Player::setData(const std::string& dataKey)
{
	ResourceSet* rs = _resman->getData(dataKey);
	if (rs == NULL)
	{
		std::string msg = cocos2d::CCString::createWithFormat("Not found data > %s", dataKey.c_str())->m_sString;
		CCAssert(rs != NULL, msg.c_str());
	}
	
	if (_currentRs != rs)
	{
		releaseData();
		rs->retain();
		_currentRs = rs;
	}
}

void Player::releaseData()
{
	releaseAnime();
	CC_SAFE_RELEASE_NULL(_currentRs);
}


void Player::releaseAnime()
{
	releaseParts();
	CC_SAFE_RELEASE_NULL(_currentAnimeRef);
}

void Player::play(const std::string& packName, const std::string& animeName, int loop, int startFrameNo, int textureSetNo)
{
	CCAssert(_currentRs != NULL, "Not select data");
	
	AnimeRef* animeRef = _currentRs->animeCache->getReference(packName, animeName);
	if (animeRef == NULL)
	{
		std::string msg = cocos2d::CCString::createWithFormat("Not found animation > pack=%s, anime=%s", packName.c_str(), animeName.c_str())->m_sString;
		CCAssert(animeRef != NULL, msg.c_str());
	}

	play(animeRef, loop, startFrameNo, textureSetNo);
}

void Player::play(const std::string& animeName, int loop, int startFrameNo, int textureSetNo)
{
	CCAssert(_currentRs != NULL, "Not select data");

	AnimeRef* animeRef = _currentRs->animeCache->getReference(animeName);
	if (animeRef == NULL)
	{
		std::string msg = cocos2d::CCString::createWithFormat("Not found animation > anime=%s", animeName.c_str())->m_sString;
		CCAssert(animeRef != NULL, msg.c_str());
	}

	play(animeRef, loop, startFrameNo, textureSetNo);
}

void Player::play(AnimeRef* animeRef, int loop, int startFrameNo, int textureSetNo)
{
	if (_currentAnimeRef != animeRef)
	{
		CC_SAFE_RELEASE_NULL(_currentAnimeRef);

		animeRef->retain();
		_currentAnimeRef = animeRef;
		
		allocParts(animeRef->animePackData->numParts, false, textureSetNo);
//		setPartsParentage();

		_playingFrame = static_cast<float>(startFrameNo);
		_step = 1.0f;
		_loop = loop;
		_loopCount = 0;
		_isPlaying = true;
		_isPausing = false;
		_prevDrawFrameNo = -1;
		
		setFrame(_playingFrame);
	}
}

void Player::pause()
{
	_isPausing = true;
}

void Player::resume()
{
	_isPausing = false;
}

void Player::stop()
{
	_isPlaying = false;
}

const std::string& Player::getPlayPackName() const
{
	return _currentAnimeRef != NULL ? _currentAnimeRef->packName : s_nullString;
}

const std::string& Player::getPlayAnimeName() const
{
	return _currentAnimeRef != NULL ? _currentAnimeRef->animeName : s_nullString;
}


void Player::update(float dt)
{
	updateFrame(dt);
}

void Player::updateFrame(float dt)
{
	if (!_currentAnimeRef) return;
	
	bool playEnd = false;
	bool toNextFrame = _isPlaying && !_isPausing;
	if (toNextFrame && (_loop == 0 || _loopCount < _loop))
	{
		// フレームを進める.
		// forward frame.
		const int numFrames = _currentAnimeRef->animationData->numFrames;

		float fdt = _frameSkipEnabled ? dt : cocos2d::CCDirector::sharedDirector()->getAnimationInterval();
		float s = fdt / (1.0f / _currentAnimeRef->animationData->fps);
		
		//if (!m_frameSkipEnabled) CCLOG("%f", s);
		
		float next = _playingFrame + (s * _step);

		int nextFrameNo = static_cast<int>(next);
		float nextFrameDecimal = next - static_cast<float>(nextFrameNo);
		int currentFrameNo = static_cast<int>(_playingFrame);
		
		if (_step >= 0)
		{
			// 順再生時.
			// normal plays.
			for (int c = nextFrameNo - currentFrameNo; c; c--)
			{
				int incFrameNo = currentFrameNo + 1;
				if (incFrameNo >= numFrames)
				{
					// アニメが一巡
					// turned animation.
					_loopCount += 1;
					if (_loop && _loopCount >= _loop)
					{
						// 再生終了.
						// play end.
						playEnd = true;
						break;
					}
					
					incFrameNo = 0;
				}
				currentFrameNo = incFrameNo;

			}
		}
		else
		{
			// 逆再生時.
			// reverse play.
			for (int c = currentFrameNo - nextFrameNo; c; c--)
			{
				int decFrameNo = currentFrameNo - 1;
				if (decFrameNo < 0)
				{
					// アニメが一巡
					// turned animation.
					_loopCount += 1;
					if (_loop && _loopCount >= _loop)
					{
						// 再生終了.
						// play end.
						playEnd = true;
						break;
					}
				
					decFrameNo = numFrames - 1;
				}
				currentFrameNo = decFrameNo;
				
			}
		}
		
		_playingFrame = static_cast<float>(currentFrameNo) + nextFrameDecimal;
	}

	setFrame(getFrameNo());
	
	if (playEnd)
	{
		stop();
	}
}




void Player::allocParts(int numParts, bool useCustomShaderProgram, int textureSetNo)
{
	releaseParts();
	setPartsParentage(textureSetNo);
}

void Player::releaseParts()
{
	removeAllChildrenWithCleanup(true);
	_parts.removeAllObjects();
}

void Player::setPartsParentage(int textureSetNo)
{
	if (!_currentAnimeRef) return;

	ToPointer ptr(_currentRs->data);
	const AnimePackData* packData = _currentAnimeRef->animePackData;
	const PartData* parts = static_cast<const PartData*>(ptr(packData->parts));
	int globalZOrder = getZOrder();
	
	cocos2d::CCTexture2D* commonTex = NULL;
	cocos2d::CCSpriteBatchNode* batchNode = NULL;
	if (_useSpriteBatchNode)
	{
		CellRef* cellRef = _currentRs->cellCache->getReference(0);
		commonTex = (cocos2d::CCTexture2D*)cellRef->textureSetRef->textures.objectAtIndex(textureSetNo);
		//this->setTexture(commonTex);
		//this->setTextureRect(cocos2d::CCRectZero);
		
		batchNode = cocos2d::CCSpriteBatchNode::createWithTexture(commonTex);
		this->addChild(batchNode);
	}

	for (int partIndex = 0; partIndex < packData->numParts; partIndex++)
	{
		const PartData* partData = &parts[partIndex];
		
		CustomSprite* sprite =  CustomSprite::create();
		sprite->_textureSetNo = textureSetNo;
		if (_useSpriteBatchNode)
		{
			sprite->setTexture(commonTex);
		}
		if (globalZOrder != 0) sprite->setZOrder(globalZOrder);
		_parts.addObject(sprite);
		
		if (partIndex > 0)
		{
			CustomSprite* parent = static_cast<CustomSprite*>(_parts.objectAtIndex(partData->parentIndex));
			sprite->_parent = parent;
			parent->addChild(sprite);
		}
		else
		{
			sprite->_parent = NULL;
			if (_useSpriteBatchNode)
			{
				batchNode->addChild(sprite);
			}
			else
			{
				this->addChild(sprite);
			}
		}
	}
}

void Player::setPartVisible(const std::string& partName, bool visibled)
{
	if (!_currentAnimeRef) return;
	
	PartsInfo::PartIndex* info = (PartsInfo::PartIndex*)_currentAnimeRef->partsInfo->indices.objectForKey(partName);
	if (info != NULL)
	{
		int partIndex = info->index;
		CustomSprite* sprite = static_cast<CustomSprite*>(_parts.objectAtIndex(partIndex));
		
		sprite->_isPartVisibled = visibled;
	}
}

void Player::setPartTextureSet(const std::string& partName, int textureSetNo)
{
	if (!_currentAnimeRef) return;
	
	PartsInfo::PartIndex* info = (PartsInfo::PartIndex*)_currentAnimeRef->partsInfo->indices.objectForKey(partName);
	if (info != NULL)
	{
		int partIndex = info->index;
		CustomSprite* sprite = static_cast<CustomSprite*>(_parts.objectAtIndex(partIndex));
		
		sprite->_textureSetNo = textureSetNo;
	}
}

void Player::setAllPartsTextureSet(int textureSetNo)
{
	if (!_currentAnimeRef) return;
	
	CCObject* part;
	CCARRAY_FOREACH(&_parts, part)
	{
		CustomSprite* sprite = static_cast<CustomSprite*>(part);
		
		sprite->_textureSetNo = textureSetNo;
	}
}

void Player::useSpriteBatchNode(bool enabled)
{
	_useSpriteBatchNode = enabled;
}

void Player::setFlippedX(bool flippedX)
{
	_flippedX = flippedX;
}

void Player::setFlippedY(bool flippedY)
{
	_flippedY = flippedY;
}

bool Player::isFlippedX(void) const
{
    return _flippedX;
}

bool Player::isFlippedY(void) const
{
    return _flippedY;
}





enum {
	PART_FLAG_INVISIBLE			= 1 << 0,
	PART_FLAG_FLIP_H			= 1 << 2,
	PART_FLAG_FLIP_V			= 1 << 3,
};

void Player::setFrame(int frameNo)
{
	if (!_currentAnimeRef) return;

	_prevDrawFrameNo = frameNo;


	ToPointer ptr(_currentRs->data);

	const AnimePackData* packData = _currentAnimeRef->animePackData;
	const PartData* parts = static_cast<const PartData*>(ptr(packData->parts));

	const AnimationData* animeData = _currentAnimeRef->animationData;
	const ss_offset* frameDataIndex = static_cast<const ss_offset*>(ptr(animeData->frameData));
	
	const ss_u16* frameDataArray = static_cast<const ss_u16*>(ptr(frameDataIndex[frameNo]));
	DataArrayReader reader(frameDataArray);


	for (int index = 0; index < packData->numParts; index++)
	{
		int partIndex = reader.readS16();
		const PartData* partData = &parts[partIndex];

		// optional parameters
		int flags      = reader.readU16();
		int cellIndex  = reader.readS16();
		int x          = reader.readS16();
		int y          = reader.readS16();
		int opacity    = reader.readU16();

		float anchorX  = reader.readFloat();
		float anchorY  = reader.readFloat();
		float rotation = -reader.readFloat();
		float scaleX   = reader.readFloat();
		float scaleY   = reader.readFloat();
		
		bool isVisibled = !(flags & PART_FLAG_INVISIBLE);
		
		CustomSprite* sprite = static_cast<CustomSprite*>(_parts.objectAtIndex(partIndex));

		if (partIndex == 0)
		{
			// rootパーツは常に座標(0,0)
			x = 0;
			y = 0;
			rotation = 0;
			// flipはrootパーツに反映する
			scaleX = isFlippedX() ? -1.0f : 1.0f;
			scaleY = isFlippedY() ? -1.0f : 1.0f;
		}
		else
		{
			// 親のアンカー位置を子の原点とするためオフセット値を足す
			CustomSprite* parent = static_cast<CustomSprite*>(_parts.objectAtIndex(partData->parentIndex));
			cocos2d::CCSize size = parent->getTextureRect().size;
			cocos2d::CCPoint anchorPt = parent->getAnchorPoint();
			float ofx = size.width * anchorPt.x;
			float ofy = size.height * anchorPt.y;
			x += ofx;
			y += ofy;
		}

		sprite->setVisible(isVisibled && sprite->_isPartVisibled);
		
		sprite->setPosition(cocos2d::CCPoint(x, y));
		sprite->setRotation(rotation);
		sprite->setScaleX(scaleX);
		sprite->setScaleY(scaleY);

		CellRef* cellRef = cellIndex >= 0 ? _currentRs->cellCache->getReference(cellIndex) : NULL;
		if (cellRef)
		{
			cocos2d::CCTexture2D* tex = (cocos2d::CCTexture2D*)cellRef->textureSetRef->textures.objectAtIndex(sprite->_textureSetNo);
			sprite->setTexture(tex);
			sprite->setTextureRect(cellRef->rect);
			
			anchorX = cellRef->cell->pivotX + anchorX;
			anchorY = cellRef->cell->pivotY + anchorY;
			sprite->setAnchorPoint(cocos2d::CCPoint(anchorX , anchorY));
		}
		else
		{
			cellRef = _currentRs->cellCache->getReference(0);
			cocos2d::CCTexture2D* tex = (cocos2d::CCTexture2D*)cellRef->textureSetRef->textures.objectAtIndex(sprite->_textureSetNo);
			sprite->setTexture(tex);
			sprite->setTextureRect(cocos2d::CCRectZero);
		}

		sprite->setFlipX(flags & PART_FLAG_FLIP_H);
		sprite->setFlipY(flags & PART_FLAG_FLIP_V);
		sprite->setOpacity(opacity);
	}
}






/**
 * CustomSprite
 */


CustomSprite::CustomSprite()
{}

CustomSprite::~CustomSprite()
{}

CustomSprite* CustomSprite::create()
{
	CustomSprite *pSprite = new CustomSprite();
	if (pSprite && pSprite->init())
	{
		pSprite->initState();
		pSprite->autorelease();
		return pSprite;
	}
	CC_SAFE_DELETE(pSprite);
	return NULL;
}




};
