using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;


public class Sbd
{
	public class ProjectData
	{
		public uint			dataId;
		public uint			version;
		public uint			flags;
		public string		imageBaseDir;
		//ss_offset cells;			// const Cell*
		//ss_offset animePacks;		// const AnimePackData*
		public int			numCells;
		public int			numAnimePacks;
		
		public Cell[]		cells;
		public AnimePackData[] animePacks;
	}
	
	public class Cell
	{
		public string		name;			// const char*
		public CellMap		cellMap;		// const CellMap*
		public int			indexInCellMap;
		public int			x;
		public int			y;
		public int			width;
		public int			height;
		//ss_s16 reserved;
		public float		pivotX;			// 0.0:left, 0.5:center, 1.0:right
		public float		pivotY;			// 0.0:bottom, 0.5:center, 1.0:top
	}
	
	public class CellMap
	{
		public string		name;			// const char*
		public string		imagePath;		// const char*
		public int			index;
		//ss_s16 reserved;
	}
	
	public class AnimePackData
	{
		public string		name;			// const char*
		//ss_offset parts;			// const PartData*
		//ss_offset animations;		// const AnimationData*
		public int			numParts;
		public int			numAnimations;
		
		public PartData[] parts;
		public AnimationData[] animations;
	}
	
	public class PartData
	{
		public string		name;			// const char*
		public int			index;
		public int			parentIndex;
		public int			type;
		public int			boundsType;
		public int			alphaBlendType;	// BlendType
	}
	
	public class AnimationData
	{
		public string		name;			// const char*
		//ss_offset frameData;		// const ss_s16*
		public int			numFrames;
		public int			fps;
		
		public FrameData[][] frameData;
	}
	
	public struct FrameData
	{
		public short index;
		public ushort flags;
		public short cellIndex;
		
		public short x;
		public short y;
		
		public short opacity;
		
		public float pivotX;
		public float pivotY;
		public float rotationZ;
		public float scaleX;
		public float scaleY;
	}
}



public class ResourceSet
{
	public Sbd.ProjectData data;
	public CellCache cellCache;
	public AnimeCache animeCache;
}


public class ResourceManager
{
	private Dictionary<string, ResourceSet> _dataDic = new Dictionary<string, ResourceSet>();
		
	public ResourceManager ()
	{
	}
	
	
	public void AddData(string name, string imageBaseDir = null)
	{
		var textAsset = Resources.Load ( name ) as TextAsset;
		if (textAsset == null) throw new UnityException("Cant load: " + name);
		byte[] rawData = textAsset.bytes;
		
		var data = Load(rawData);
		if (data == null) throw new UnityException("Failure parse data: " + name);

		if (imageBaseDir == null)
		{
			imageBaseDir = Path.GetDirectoryName(name);
		}

		var rs = new ResourceSet();
		rs.data = data;
		rs.cellCache = new CellCache(data.cells, imageBaseDir);
		rs.animeCache = new AnimeCache(data);

		_dataDic[name] = rs;
	}


	public void RemoveAllData()
	{
		_dataDic.Clear();
	}


	public void ChangeCell(string name, string cellName, string changeTextureName, int x, int y)
	{
		if (_dataDic.ContainsKey(name))
		{
			var resSet = _dataDic[name];
			resSet.cellCache.ChangeCell(cellName, changeTextureName, x, y);
		}
	}


	public ResourceSet GetData(string name)
	{
		return _dataDic[name];
	}


	private static readonly ResourceManager s_instance = new ResourceManager();
	
	public static ResourceManager Instance()
	{
		return s_instance;
	}
	


	public static Sbd.ProjectData Load(byte[] rawBytes)
	{
		var reader = new DataReader(rawBytes);
		
		var data = new Sbd.ProjectData();
		data.dataId = reader.ReadUInt32();
		data.version = reader.ReadUInt32();
		data.flags = reader.ReadUInt32();
		data.imageBaseDir = reader.ReadString();
		
		var cellReader = reader.ReadData();
		var animePackReader = reader.ReadData();
		
		data.numCells = reader.ReadInt16();
		data.numAnimePacks = reader.ReadInt16();
		
		data.cells = ReadCells(cellReader, data.numCells);
		data.animePacks = ReadAnimePackData(animePackReader, data.numAnimePacks);
		
		return data;
	}
	
	
	
	static Sbd.Cell[] ReadCells(DataReader reader, int numCells)
	{
		var cellMapDic = new Dictionary<uint, Sbd.CellMap>();
		
		var cells = new Sbd.Cell[numCells];
		for (int i = 0; i < numCells; i++ )
		{
			var cell = new Sbd.Cell();
			cell.name = reader.ReadString();
			
			var cellMapOffset = reader.ReadUInt32();
			Sbd.CellMap cellMap;
			if (!cellMapDic.TryGetValue(cellMapOffset, out cellMap))
			{
				var cellMapReader = new DataReader(reader.Data, cellMapOffset);
				cellMap = ReadCellMap(cellMapReader);
				cellMapDic[cellMapOffset] = cellMap;
			}
			cell.cellMap = cellMap;
			
			cell.indexInCellMap = reader.ReadInt16();
			cell.x = reader.ReadInt16();
			cell.y = reader.ReadInt16();
			cell.width = reader.ReadInt16();
			cell.height = reader.ReadInt16();
			reader.ReadInt16();					// data skip;
			cell.pivotX = reader.ReadSingle();
			cell.pivotY = reader.ReadSingle();
			
			cells[i] = cell;
		}
		return cells;
	}
	
	
	static Sbd.CellMap ReadCellMap(DataReader reader)
	{
		var cellMap = new Sbd.CellMap();
		cellMap.name = reader.ReadString();
		cellMap.imagePath = reader.ReadString();
		cellMap.index = reader.ReadInt16();
		return cellMap;
	}
	
	
	static Sbd.AnimePackData[] ReadAnimePackData(DataReader reader, int numAnimePacks)
	{
		var animePacks = new Sbd.AnimePackData[numAnimePacks];
		for (int i = 0; i < numAnimePacks; i++)
		{
			var animePack = new Sbd.AnimePackData();
			animePack.name = reader.ReadString();
			
			var partReader = reader.ReadData();
			var animationReader = reader.ReadData();
			
			animePack.numParts = reader.ReadInt16();
			animePack.numAnimations = reader.ReadInt16();
			
			animePack.parts = ReadPartData(partReader, animePack.numParts);
			animePack.animations = ReadAnimationData(animationReader, animePack.numAnimations, animePack.numParts);
			
			animePacks[i] = animePack;
		}
		return animePacks;
	}
	
	
	static Sbd.PartData[] ReadPartData(DataReader reader, int numParts)
	{
		var parts = new Sbd.PartData[numParts];
		for (int i = 0; i < numParts; i++)
		{
			var part = new Sbd.PartData();
			part.name = reader.ReadString();
			part.index = reader.ReadInt16();
			part.parentIndex = reader.ReadInt16();
			part.type = reader.ReadInt16();
			part.boundsType = reader.ReadInt16();
			part.alphaBlendType = reader.ReadInt16();
			reader.ReadInt16();	// skip data
			
			parts[i] = part;
		}
		return parts;
	}
	
	
	static Sbd.AnimationData[] ReadAnimationData(DataReader reader, int numAnimations, int numParts)
	{
		var animations = new Sbd.AnimationData[numAnimations];
		for (int i = 0; i < numAnimations; i++)
		{
			var animation = new Sbd.AnimationData();
			animation.name = reader.ReadString();
			
			var frameDataIndexReader = reader.ReadData();
			
			animation.numFrames = reader.ReadInt16();
			animation.fps = reader.ReadInt16();
			
			animation.frameData = ReadFrameData(frameDataIndexReader, animation.numFrames, numParts);
			
			animations[i] = animation;
		}
		return animations;
	}
	
	
	static Sbd.FrameData[][] ReadFrameData(DataReader indexReader, int numFrames, int numParts)
	{
		var frameData = new Sbd.FrameData[numFrames][];
		for (int i = 0; i < numFrames; i++)
		{
			var reader = indexReader.ReadData();
			
			frameData[i] = new Sbd.FrameData[numParts];
			for (int p = 0; p < numParts; p++)
			{
				frameData[i][p].index = reader.ReadInt16();
				frameData[i][p].flags = reader.ReadUInt16();
				frameData[i][p].cellIndex = reader.ReadInt16();
				
				frameData[i][p].x = reader.ReadInt16();
				frameData[i][p].y = reader.ReadInt16();
				
				frameData[i][p].opacity = reader.ReadInt16();
				
				frameData[i][p].pivotX = reader.ReadSingle();
				frameData[i][p].pivotY = reader.ReadSingle();
				frameData[i][p].rotationZ = reader.ReadSingle();
				frameData[i][p].scaleX = reader.ReadSingle();
				frameData[i][p].scaleY = reader.ReadSingle();
			}
		}
		return frameData;
	}

}



public class CellRef
{
	public Sbd.Cell cell;
	public Sprite[] sprites;
}


public class CellCache
{
	private Dictionary<string, Texture2D[]> _textureDic = new Dictionary<string, Texture2D[]>();
	private CellRef[] _cellRefs;
	
	
	public CellCache(Sbd.Cell[] cells, string imageBaseDir = null)
	{
		var cellRefs = new List<CellRef>(cells.Length);
		foreach (var cell in cells)
		{
			var texSet = CreateTexture(cell.cellMap.imagePath, imageBaseDir);

			var sprites = new List<Sprite>();
			foreach (var tex in texSet)
			{
				var texRect = new Rect(cell.x, tex.height - (cell.y + cell.height), cell.width, cell.height);
				//var sprite = Sprite.Create(tex, texRect, new Vector2(cell.pivotX, cell.pivotY), 100, 1, SpriteMeshType.FullRect);
				var sprite = Sprite.Create(tex, texRect, new Vector2(cell.pivotX, cell.pivotY), 1, 1, SpriteMeshType.FullRect);
				sprites.Add(sprite);
			}

			var cellRef = new CellRef()
			{
				cell = cell,
				sprites = sprites.ToArray(),
			};
			cellRefs.Add(cellRef);
		}
		_cellRefs = cellRefs.ToArray();
	}
	

	Texture2D[] CreateTexture(string imagePath, string imageBaseDir)
	{
		var texPath = Path.GetFileNameWithoutExtension(imagePath);
		if (imageBaseDir != null)
		{
			//texPath = Path.Combine(imageBaseDir, texPath);
			texPath = imageBaseDir + "/" + texPath;
		}

		var dirname = Path.GetDirectoryName(texPath);
		var filename = Path.GetFileNameWithoutExtension(texPath);

		if (!_textureDic.ContainsKey(filename))
		{
			var basename = filename.Substring(0, filename.Length - 1);

			var texSet = new List<Texture2D>();
			for (int i = 1; i <= 2; i++)
			{
				//var loadTexPath = Path.Combine(dirname, basename + i);
				var loadTexPath = dirname + "/" + basename + i;
				var tex = Resources.Load(loadTexPath) as Texture2D;
				if (tex == null) throw new UnityException("Failed texture load: " + texPath);
				//Debug.Log("Load: " + loadTexPath);
				texSet.Add(tex);
			}

			_textureDic.Add(filename, texSet.ToArray());
		}
		return _textureDic[filename];
	}
	
	
	public CellRef GetReference(int index)
	{
		if (index < 0 || index >= _cellRefs.Length) throw new UnityException();
		return _cellRefs[index];
	}
	

	public void ChangeCell(string cellName, string changeTextureName, int x, int y)
	{
		var cellRef = _cellRefs.FirstOrDefault(r => r.cell.name == cellName);
		if (cellRef == null) return;

		if (!_textureDic.ContainsKey(changeTextureName)) return;
		var texSet = _textureDic[changeTextureName];
		var cell = cellRef.cell;

		var sprites = new List<Sprite>();
		foreach (var tex in texSet)
		{
			var texRect = new Rect(x, tex.height - (y + cell.height), cell.width, cell.height);
			//var sprite = Sprite.Create(tex, texRect, new Vector2(cell.pivotX, cell.pivotY), 100, 1, SpriteMeshType.FullRect);
			var sprite = Sprite.Create(tex, texRect, new Vector2(cell.pivotX, cell.pivotY), 1, 1, SpriteMeshType.FullRect);
			sprites.Add(sprite);
		}

		cellRef.sprites = sprites.ToArray();
	}
}



public class AnimeRef
{
	public string packName;
	public string animeName;
	public Sbd.AnimationData animationData;
	public Sbd.AnimePackData animePackData;

	public Dictionary<string, int> partIndices;
}


public class AnimeCache
{
	private Dictionary<string, AnimeRef> _animeDic = new Dictionary<string, AnimeRef>();


	public AnimeCache(Sbd.ProjectData data)
	{
		foreach (var pack in data.animePacks)
		{
			var partIndices = new Dictionary<string, int>();
			foreach (var i in pack.parts.Select((part, index) => new { part, index }))
			{
				partIndices[i.part.name] = i.index;
			}

			foreach (var animation in pack.animations)
			{
				var animeRef = new AnimeRef()
				{
					packName = pack.name,
					animeName = animation.name,
					animationData = animation,
					animePackData = pack,
					partIndices = partIndices,
				};


				// register is packName + animeName
				var key = ToPackAnimeKey(pack.name, animation.name);
				_animeDic[key] = animeRef;

				// register is animeName only
				_animeDic[animation.name] = animeRef;
			}
		}
	}


	public AnimeRef GetReference(string packName, string animeName)
	{
		var key = ToPackAnimeKey(packName, animeName);
		return _animeDic[key];
	}


	public AnimeRef GetReference(string animeName)
	{
		return _animeDic[animeName];
	}


	static string ToPackAnimeKey(string packName, string animeName)
	{
		return packName + "/" + animeName;
	}
}



public class DataReader
{
	private byte[] _data;
	private BinaryReader _reader;
	
	
	public DataReader(byte[] data, uint offset = 0)
	{
		_data = data;
		_reader = CreateReader(offset);
	}
	
	private BinaryReader CreateReader(uint offset)
	{
		int index = (int)offset;
		var stream = new MemoryStream(_data, index, _data.Length - index);
		return new BinaryReader(stream);
	}
	
	
	public byte[] Data
	{
		get { return _data; }
	}
	
	public uint ReadUInt32()
	{
		return _reader.ReadUInt32();
	}
	
	public short ReadInt16()
	{
		return _reader.ReadInt16();
	}
	
	public ushort ReadUInt16()
	{
		return _reader.ReadUInt16();
	}
	
	public float ReadSingle()
	{
		return _reader.ReadSingle();
	}
	
	public string ReadString()
	{
		var offset = _reader.ReadUInt32();
		if (offset != 0)
		{
			var reader = CreateReader(offset);
			var sb = new StringBuilder();
			for (char ch; (ch = reader.ReadChar()) != 0; )
			{
				sb.Append(ch);
			}
			return sb.ToString();
		}
		else
		{
			return null;
		}
	}
	
	public DataReader ReadData()
	{
		var offset = _reader.ReadUInt32();
		return new DataReader(_data, offset);
	}
}


