using UnityEngine;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

public class SbdPlayer : MonoBehaviour
{

	public delegate void PlayEndHandler();

	public event PlayEndHandler PlayEnd;



	public static SbdPlayer Create(string name)
	{
		var obj = new GameObject(name);
		obj.AddComponent<SbdPlayer>();
		return obj.GetComponent<SbdPlayer>();
	}




	public class PartInfo
	{
		public int	textureSetNo { get; set; }
	}



	ResourceManager _resman;
	ResourceSet _currentRs;
	AnimeRef _currentAnimeRef;
	GameObject[] _parts;
	PartInfo[] _partInfos;

	float _playingFrame;


	public SbdPlayer ()
	{
		_resman = ResourceManager.Instance();
	}


	void Start () {

	}
	

	void Update () {
		if (_currentAnimeRef == null) return;

		float fdt = Time.deltaTime;
		float s = fdt / (1.0f / _currentAnimeRef.animationData.fps);
		
		float nextFrame = _playingFrame + s;
		
		int frameNo = (int)nextFrame;
		if (frameNo >= _currentAnimeRef.animationData.numFrames)
		{
			nextFrame = 0.0f;
			frameNo = 0;

			if (PlayEnd != null)
			{
				PlayEnd();
			}
		}
		_playingFrame = nextFrame;
		
		SetFrame(frameNo);
	}


	public void SetData(string dataKey)
	{
		var rs = _resman.GetData(dataKey);
		if (rs == null) throw new ArgumentException("Not found data: " + dataKey);

		if (_currentRs != rs)
		{
			ReleaseParts();
			_currentRs = rs;
		}
	}


	public void Play(string animeName, int loop = 0, int startFrameNo = 0)
	{
		if (animeName == null) throw new ArgumentException();

		var animeRef = _currentRs.animeCache.GetReference(animeName);
		if (animeRef == null) throw new ArgumentException("Not found animation: " + animeName);

		Play (animeRef, loop, startFrameNo);
	}


	void Play(AnimeRef animeRef, int loop, int startFrameNo)
	{
		if (_currentAnimeRef != animeRef)
		{
			_currentAnimeRef = animeRef;

			AllocParts(animeRef.animePackData.parts);
			SetRelation(animeRef.animePackData.parts);

			_playingFrame = (float)startFrameNo;

			SetFrame(startFrameNo);
		}
	}


	void AllocParts(Sbd.PartData[] partList)
	{
		if (_parts != null) return;

		var parts = new List<GameObject>();
		foreach (var part in partList.Reverse())
		{
			var obj = new GameObject (part.name);
			obj.AddComponent<SpriteRenderer>();

			parts.Add(obj);
		}
		parts.Reverse();
		_parts = parts.ToArray();

		var infos = new List<PartInfo>();
		foreach (var part in parts)
		{
			infos.Add(new PartInfo());
		}
		_partInfos = infos.ToArray();
	}


	void SetRelation(Sbd.PartData[] partList)
	{
		foreach (var part in partList.Reverse())
		{
			var obj = _parts[part.index];
			if (part.parentIndex >= 0)
			{
				var parentObj = _parts[part.parentIndex];
				obj.transform.parent = parentObj.transform;
			}
			else
			{
				obj.transform.parent = this.transform;
			}
		}
	}


	void ReleaseParts()
	{
		if (_parts != null)
		{
			foreach (var part in _parts.Reverse())
			{
				Destroy(part);
			}
			_parts = null;
		}
	}


	void SetFrame(int frameNo)
	{
		if (_currentAnimeRef == null) return;
		
		var frameData = _currentAnimeRef.animationData.frameData[frameNo];
		for (int partNo = 0; partNo < _parts.Length; partNo++)
		{
			var obj = _parts[partNo];
			var info = _partInfos[partNo];

			SpriteRenderer renderer = obj.GetComponent<SpriteRenderer>();

			Sprite sprite = null;
			if (frameData[partNo].cellIndex >= 0)
			{
				var cellRef = _currentRs.cellCache.GetReference(frameData[partNo].cellIndex);
				sprite = cellRef.sprites[info.textureSetNo];
			}
			else
			{
				var cellRef = _currentRs.cellCache.GetReference(0);
				sprite = cellRef.sprites[info.textureSetNo];
				renderer.enabled = false;
			}

			if (renderer.sprite != sprite)
			{
				renderer.sprite = sprite;
			}

			//obj.transform.localPosition = new Vector3((float)frameData[partNo].x / 100, (float)frameData[partNo].y / 100, 0);
			obj.transform.localPosition = new Vector3((float)frameData[partNo].x, (float)frameData[partNo].y, 0);
			obj.transform.localRotation = Quaternion.Euler(new Vector3(0, 0, frameData[partNo].rotationZ));
			obj.transform.localScale = new Vector3(frameData[partNo].scaleX, frameData[partNo].scaleY, 1);
		}
	}


	public void SetPartVisible(string partName, bool isVisibled)
	{
		if (_currentAnimeRef == null) return;

		if (_currentAnimeRef.partIndices.ContainsKey(partName))
		{
			var obj = _parts[_currentAnimeRef.partIndices[partName]];
			obj.renderer.enabled = isVisibled;
//			obj.SetActive(isVisibled);
//			SetRelation(_currentAnimeRef.animePackData.parts);
		}
	}


	public void SetPartTextureSet(string partName, int textureSetNo)
	{
		if (_currentAnimeRef == null) return;
		
		if (_currentAnimeRef.partIndices.ContainsKey(partName))
		{
			var info = _partInfos[_currentAnimeRef.partIndices[partName]];
			info.textureSetNo = textureSetNo;
		}
	}


	public void SetAllPartsTextureSet(int textureSetNo)
	{
		if (_currentAnimeRef == null) return;

		foreach (var info in _partInfos)
		{
			info.textureSetNo = textureSetNo;
		}
	}

}

